use crate::prelude::*;

pub(crate) struct Args {
    pub(crate) port: u16,
}

impl Default for Args {
    fn default() -> Self {
        Self { port: 9000 }
    }
}

pub(crate) struct App {
    client_streams: BTreeMap<Id, Framed<TcpStream, LengthDelimitedCodec>>,
    listener: TcpListener,
    server: Server,
}

impl App {
    pub(crate) async fn new(args: Args) -> Result<Self> {
        let listener = TcpListener::bind(("0.0.0.0", args.port)).await?;

        Ok(Self {
            client_streams: Default::default(),
            listener,
            server: Default::default(),
        })
    }

    pub(crate) fn poll_run(&mut self, cx: &mut Context<'_>) -> Poll<Result<()>> {
        match self.step(cx) {
            Ok(()) => Poll::Pending,
            Err(err) => Poll::Ready(Err(err)),
        }
    }

    fn step(&mut self, cx: &mut Context<'_>) -> Result<()> {
        let mut clients_to_remove = vec![];

        // Pump the network streams
        for (id, stream) in &mut self.client_streams {
            let mut stream = pin!(stream);

            // Try to send from the client's outbox
            if let Poll::Ready(result) =
                <_ as futures_sink::Sink<Bytes>>::poll_ready(stream.as_mut(), cx)
            {
                result.context("Can't check network write half for readiness")?;
                if let Some(frame) = self.server.poll_send(*id)? {
                    stream.as_mut().start_send(frame).context("start_send")?;
                    if let Poll::Ready(result) =
                        <_ as futures_sink::Sink<Bytes>>::poll_flush(stream.as_mut(), cx)
                    {
                        result.context("poll_flush")?;
                    }
                    tracing::debug!("Started send");
                }
            }

            // Try to read data in
            match stream.as_mut().poll_next(cx) {
                Poll::Pending => {}
                Poll::Ready(None) => clients_to_remove.push(*id),
                Poll::Ready(Some(result)) => {
                    let frame = result.context("stream.poll_next")?;
                    cx.waker().wake_by_ref();
                    self.server.handle_client_frame(*id, frame.into())?;
                }
            }
        }

        // Close out disconnected clients
        for id in clients_to_remove {
            cx.waker().wake_by_ref();
            tracing::info!(?id, "Closing client");
            self.client_streams.remove(&id);
            self.server.handle_client_disconnected(id)?;
        }

        if let Poll::Ready(result) = self.listener.poll_accept(cx) {
            let (stream, _addr) = result.context("listener.poll_accept")?;
            cx.waker().wake_by_ref();
            let stream = Framed::new(stream, LengthDelimitedCodec::new());
            let id = self.server.handle_new_client()?;
            self.client_streams.insert(id, stream);
        }

        Ok(())
    }
}

#[derive(Default)]
struct Server {
    clients: BTreeMap<Id, Client>,
    next_client_id: Id,
}

impl Server {
    fn broadcast(&mut self, msg: &ToClient) -> Result<()> {
        for client in &mut self.clients.values_mut() {
            client.handle_outgoing(msg)?;
        }
        Ok(())
    }

    fn handle_client_disconnected(&mut self, id: Id) -> Result<()> {
        let client = self
            .clients
            .remove(&id)
            .context("Can't remove client who isn't there")?;
        self.broadcast(&ToClient {
            id,
            name: client.name,
            event: ToClientEvent::Disconnected,
        })?;
        Ok(())
    }

    fn handle_client_frame(&mut self, id: Id, frame: Bytes) -> Result<()> {
        let (name, event) = {
            let client = self
                .clients
                .get_mut(&id)
                .context("Logic error: Stream has no associated client")?;
            let Some(event) = client.handle_frame(frame).context("client.handle_frame")? else {
                return Ok(());
            };
            (client.name.clone(), event)
        };
        self.broadcast(&ToClient { id, name, event })?;
        Ok(())
    }

    fn handle_new_client(&mut self) -> Result<Id> {
        let id = self.next_client_id;
        self.next_client_id += 1;
        let client = Client {
            id,
            name: "anon".to_string(),
            outbox: Default::default(),
            sequence: 0,
        };
        tracing::info!(
            id = client.id,
            total = self.clients.len(),
            "Accepted client"
        );
        self.broadcast(&ToClient {
            id,
            name: client.name.clone(),
            event: ToClientEvent::Connected,
        })?;
        self.clients.insert(id, client);
        Ok(id)
    }

    fn poll_send(&mut self, id: Id) -> Result<Option<Bytes>> {
        let client = self
            .clients
            .get_mut(&id)
            .context("Logic error: Stream has no associated client")?;
        Ok(client.poll_send())
    }
}

struct Client {
    id: u64,
    name: String,
    outbox: VecDeque<Bytes>,
    sequence: u64,
}

impl Client {
    fn handle_frame(&mut self, frame: Bytes) -> Result<Option<ToClientEvent>> {
        match rmp_serde::from_slice(&frame)? {
            ToServer::ChatLine { line, sequence } => {
                if sequence != self.sequence {
                    tracing::error!(
                        expected = self.sequence,
                        actual = sequence,
                        "Sequence mismatch"
                    );
                    bail!("Sequence mismatch");
                }
                self.sequence += 1;
                Ok(Some(ToClientEvent::ChatLine { line }))
            }
            ToServer::SetName { name } => {
                let old_name = std::mem::replace(&mut self.name, name);
                Ok(Some(ToClientEvent::NameChanged { old_name }))
            }
        }
    }

    fn handle_outgoing(&mut self, msg: &ToClient) -> Result<()> {
        if !self.outbox.is_empty() {
            bail!("Outbox full");
        }
        let bytes = rmp_serde::to_vec(msg)?.into();
        self.outbox.push_back(bytes);
        Ok(())
    }

    fn poll_send(&mut self) -> Option<Bytes> {
        self.outbox.pop_front()
    }
}
